packages <- c(
  "cowplot",
  "shiny",
  "magrittr",
  "raster",
  "data.table",
  "MASS",
  "ggplot2",
  "ggsci"
)
new_packages <- packages[
  !(packages %in% installed.packages()[, "Package"])
]
if (length(new_packages)) install.packages(new_packages)

library("magrittr")
library("raster")
library("data.table")
library("MASS")
library("ggplot2")
library("ggsci")

#' Helper function for easy conditional on data tables.
iff <- function(
  olddat,
  condition,
  newdat
) {
  if (condition) {
    return(newdat)
  } else {
    return(olddat)
  }
}

#' Read tif files into R by providing a path.
#' Include XY includes the position coordinates
read_tif <- function(
  path,
  exclude_zero = TRUE,
  include_xy = FALSE
) {
  x <- path %>% raster()
  dat <- data.table(
    y = nrow(x) - rep(seq_len(nrow(x)), each = ncol(x)) + 1,
    x = rep(seq_len(ncol(x)), nrow(x)),
    I = as.vector(x)
  ) %>%
  iff(exclude_zero, .[I > 0]) %>%
  iff(!include_xy, .[, I])
  return(dat)
}

#' Join 2 tif data imported using read_tif().
#' Both imports should include XY.
join_tif <- function(
  dat1,
  dat2,
  exclude_zero = TRUE,
  include_xy = FALSE
) {
  dat <- merge(dat1, dat2, by = c("x", "y")) %>%
  iff(exclude_zero, .[I.x > 0 & I.y > 0]) %>%
  iff(!include_xy, .[, .(I.x, I.y)])
  return(dat)
}

#' get a nxn matrix with joint densities.
get_joint_density <- function(
  dat,
  nbins = 25
) {
  density <- kde2d(dat$x, dat$y, n = nbins)
  dat <- data.table(
    x = rep(density$x, nbins),
    y = rep(density$y, each = nbins),
    density = as.vector(density$z)
  )
  return(dat)
}

find_density_box <- function(dat, box = 0.05) {
  result <- list()
  midx <- dat[density == max(density), unique(x)]
  midy <- dat[density == max(density), unique(y)]
  xmin <- dat[
    ,
    max(
      midx - (0.5 * box * (max(x) - min(x))),
      min(x)
    )
  ]
  xmax <- dat[
    ,
    min(
      midx + (0.5 * box * (max(x) - min(x))),
      max(x)
    )
  ]
  ymin <- dat[
    ,
    max(
      midy - (0.5 * box * (max(y) - min(y))),
      min(y)
    )
  ]
  ymax <- dat[
    ,
    min(
      midy + (0.5 * box * (max(y) - min(y))),
      max(y)
    )
  ]
  result$x <- c(xmin, xmax)
  result$y <- c(ymin, ymax)
  return(result)
}

predict_joint_density <- function(dat, nbins = 25) {
  density <- get_joint_density(dat, nbins)
  x_spline <- smooth.spline(density$x, density$density)
  y_spline <- smooth.spline(density$y, density$density)
  dat <- dat[,
    .(
      x,
      y,
      density = predict(x_spline, x)[[2]] * predict(y_spline, y)[[2]]
    )
  ]
  return(dat)
}

#' Smooth a Y data with respect to X using n bins.
smooth_data <- function(x, y, nbins = 100) {
  smooth <- smooth.spline(x, y)
  dat <- predict(
    smooth,
    seq(min(x), max(x), length.out = nbins)
  ) %>%
  as.data.table()
}

#' identify Y peaks with respect to X
find_peaks <- function(x, y, npeaks = 1) {
  dat <- data.table()
  sample_width <- max(1, 0.025 * length(x))
  s <- smooth_data(x, y, length(x)) %>% .[, y]
  peaks <- which(diff(sign(diff(s))) == -2) + 1
  for (i in seq_len(length(peaks))) {
    subset <- seq(
      max(peaks[i] - sample_width, 1),
      min(peaks[i] + sample_width, length(x))
    )
    dat <- dat %>%
    rbind(
      summarize_freq_table(
        data.table(
          x = x[subset],
          frequency = y[subset]
        )
      ) %>%
      .[, frequency := y[peaks[i]]]
    )
  }
  dat <- dat[order(-frequency)][1:npeaks]
  return(dat)
}

normalize_data <- function(x, y, mode = "area") {
  if (mode == "area") {
    bin_w <- (max(x) - min(x)) / length(x)
    y <- y / (sum(y) * bin_w)
  } else if (mode == "minmax") {
    y <- (y - min(y)) / (max(y) - min(y))
  } else {
    y <- y
  }
  return(y)
}

get_freq_table <- function(
  x,
  nbins = 25,
  normalize = "none"
) {
  h <- hist(x, breaks = nbins, plot = FALSE)
  dat <- data.table(
    x = h$mids,
    frequency = h$counts
  ) %>%
  .[, frequency := normalize_data(x, frequency, eval(normalize))]
  return(dat)
}

summarize_freq_table <- function(
  dat
) {
  dat <- dat %>%
  .[order(x)] %>%
  .[, n := sum(frequency)] %>%
  .[, cumfreq := cumsum(frequency)] %>%
  .[, quartiles := cumsum(frequency / n)] %>%
  .[, q1 := x[quartiles >= 0.25][1]] %>%
  .[, median := x[quartiles >= 0.5][1]] %>%
  .[, q3 := x[quartiles >= 0.75][1]] %>%
  .[, qmin := max(q1 - 1.5 * (q3 - q1), min(x))] %>%
  .[, qmax := min(q3 + 1.5 * (q3 - q1), max(x))] %>%
  .[, min := min(x)] %>%
  .[, max := max(x)] %>%
  .[, mean := sum(x * frequency) / n] %>%
  .[, sd := sqrt(sum(frequency * (x - mean)^2) / n)] %>%
  .[, -c("x", "frequency", "cumfreq", "quartiles")] %>%
  unique()
  return(dat)
}

resample <- function(
  dat,
  factor = 0.5
) {
  dat <- dat[sample(.N, max(1, .N * factor))]
  return(dat)
}

draw_tif <- function(
  dat
) {
  p <- dat %>%
  ggplot(aes(x = x, y = y, fill = I)) +
  geom_tile() +
  coord_fixed(1) +
  scale_fill_gradient(low = "black", high = "white") +
  theme_void() +
  labs(
    title = "", x = "", y = "", fill = "Intensity"
  )
}

draw_boxplot <- function(
  dat,
  name_column = NULL,
  show_axis = TRUE
) {
  p <- dat %>%
  ggplot(
    aes(
      y = 1,
      xmin = qmin,
      xlower = q1,
      xmiddle = median,
      xupper = q3,
      xmax = qmax
    )
  ) +
  geom_boxplot(
    aes_string(color = name_column),
    stat = "identity",
    width = 0.2,
    position = position_dodge2(1.8)
  ) +
  scale_color_npg() +
  theme_classic() +
  {if (!show_axis) theme_void()} +
  {if (!show_axis) theme(legend.position = "none")} +
  labs(x = "", y = "", title = "", color = "") +
  xlim(min(dat$min), max(dat$max))
  return(p)
}

draw_hist <- function(
  dat,
  name_column = NULL,
  peaks = 0
) {
  p <- dat %>%
  ggplot(aes(x = x, y = frequency)) +
  geom_line(
    aes_string(color = name_column)
  ) +
  {if (is.null(name_column)) scale_color_npg()} +
  {if (!is.null(name_column)) scale_color_manual(
    breaks = c(
      grep("Overall", unique(dat[[name_column]]), invert = TRUE, value = TRUE),
      "Overall"
    ),
    values = c(
      pal_npg("nrc")(
        length(
          grep(
            "Overall",
            unique(dat[[name_column]]),
            invert = TRUE,
            value = TRUE
          )
        )
      ),
      "black"
    )
  )} +
  {if (peaks > 0) mapply(
    function(mean, sd, scale) {
      stat_function(
        fun = function(x, mean, sd, scale) {
          (scale * dnorm(x, mean, sd)) / dnorm(mean, mean, sd)
        },
        args = list(mean = mean, sd = sd, scale = scale),
        linetype = 2,
        color = "red"
      )
    },
    mean = dat[, .(find_peaks(x, frequency, eval(peaks)))][, mean],
    sd = dat[, .(find_peaks(x, frequency, eval(peaks)))][, sd],
    scale = dat[, .(find_peaks(x, frequency, eval(peaks)))][, frequency]
  )} +
  {if (peaks > 0) labs(
    caption = dat[
      ,
      .(find_peaks(x, frequency, eval(peaks)))
    ][
      ,
      paste("Peak", seq_len(.N), "[", "Mean:", round(mean, 2), "SD:", round(sd, 2), "]")
    ] %>%
    paste(collapse = "\n")
  )} +
  theme_classic() +
  labs(x = "", y = "", title = "", color = "") +
  xlim(min(dat$x), max(dat$x))
  return(p)
}

draw_scatter <- function(
  dat,
  normalize = FALSE,
  density_column = NULL,
  box = 0
) {
  p <- dat %>%
  ggplot(aes(x = x, y = y)) +
  geom_jitter(
    aes_string(color = density_column),
    alpha = 0.2, width = 0.5, height = 0.5
  ) +
  theme_classic() +
  xlim(min(dat$x), max(dat$x)) +
  ylim(min(dat$y), max(dat$y)) +
  {if (normalize) scale_x_continuous(
    breaks = seq(
      min(dat$x),
      max(dat$x),
      length.out = 5
    ),
    labels = seq(
      0,
      1,
      length.out = 5
    )
  )} +
  {if (normalize) scale_y_continuous(
    breaks = seq(
      min(dat$y),
      max(dat$y),
      length.out = 5
    ),
    labels = seq(
      0,
      1,
      length.out = 5
    )
  )} +
  coord_fixed(
    max(dat$x) / max(dat$y)
  ) +
  {if (!is.null(density_column)) scale_color_gradientn(
    colors = rev(terrain.colors(5)),
    breaks = seq(
      min(dat$density),
      max(dat$density),
      length.out = 5
    ),
    labels = seq(
      0,
      1,
      length.out = 5
    )
  )} +
  {if (!is.null(density_column) & box > 0) geom_rect(
    xmin = find_density_box(dat, box)[[1]][1],
    xmax = find_density_box(dat, box)[[1]][2],
    ymin = find_density_box(dat, box)[[2]][1],
    ymax = find_density_box(dat, box)[[2]][2],
    fill = NA,
    color = "black",
    linetype = 2
  )} +
  {if (!is.null(density_column) & box > 0) labs(
    caption = paste(
      "X range:",
      find_density_box(dat, box)[[1]][1] %>%
      round(2),
      "to",
      find_density_box(dat, box)[[1]][2] %>%
      round(2),
      "\nY range:",
      find_density_box(dat, box)[[2]][1] %>%
      round(2),
      "to",
      find_density_box(dat, box)[[2]][2] %>%
      round(2)
    )
  )} +
  labs(x = "", y = "", title = "", color = "")
  return(p)
}