library("shiny")
library("cowplot")
source("./functions.R")

render_files <- function(input, output) {
  output$file_selectors <- renderUI({
    files <- list()
    for (i in seq(input$nfiles)) {
      id <- paste0("file", i)
      label <- paste0("File ", i)
      files[[i]] <- fileInput(
        id,
        label,
        buttonLabel = "Browse..",
        placeholder = ifelse(
          is.null(input[[id]]),
          "No file selected",
          input[[id]]$name
        )
      )
    }
    files
  })
}
render_xy_selectors <- function(input, output) {
  output$stream_selectors <- renderUI({
    selectors <- list()
    if (input$nfiles >= 2) {
      for (i in seq_len(input$nfiles - 1)) {
        val <- paste0("file", seq_len(input$nfiles))
        names(val) <- paste("File", seq_len(input$nfiles))
        selectors[[i]] <- fluidRow(
          column(
            6,
            selectInput(
              paste0("xstream", i),
              label = paste("X Stream", i),
              choices = as.list(val),
              selected = input[[paste0("xstream", i)]]
            )
          ),
          column(
            6,
            selectInput(
              paste0("ystream", i),
              label = paste("Y Stream", i),
              choices = as.list(val),
              selected = input[[paste0("ystream", i)]]
            )
          )
        )
      }
    }
    selectors
  })
}

render_preview <- function(input, output, data_fun) {
  output$file_preview <- renderPlot({
    dat <- data_fun()
    if (length(dat) > 0) {
      plots <- list()
      for (i in seq_len(length(unique(dat$name)))) {
        plots[[i]] <- dat[name == unique(dat$name)[i]] %>%
        draw_tif() +
        labs(title = paste("Image Preview:", unique(dat$name)[i]))
      }
      if (length(plots) > 0) {
        plot_grid(plotlist = plots, nrow = 1)
      }
    }
  })
}
render_hist <- function(input, output, data_fun) {
  output$histograms <- renderPlot({
    dat <- data_fun()
    if (input$show_histogram) {
      if (length(dat) > 0) {
        dat <- dat[, .(I, name)] %>%
        iff(input$exclude_zero, .[I > 0]) %>%
        iff(
          input$combine_hist,
          .[, .(I = c(I, I), name = c(name, rep("Overall", .N)))]
        ) %>%
        iff(
          (input$color_hist & !input$show_overall),
          .[name != "Overall"]
        ) %>%
        iff(
          (!input$color_hist & input$combine_hist),
          .[name == "Overall"]
        )
        output$histogramdata <- downloadHandler(
          filename = "histogramdata.csv",
          content = function(file) {
            write.csv(
              dat[
                ,
                get_freq_table(
                  I,
                  nbins = input$histbins,
                  normalize = "area"
                ),
                name
              ],
              file,
              row.names = FALSE
            )
          }
        )
        output$boxdata <- downloadHandler(
          filename = "boxplotdata.csv",
          content = function(file) {
            write.csv(
              dat[
                ,
                get_freq_table(
                  I,
                  nbins = input$histbins,
                  normalize = "area"
                ),
                name
              ] %>%
              .[, summarize_freq_table(data.table(x, frequency)), name],
              file,
              row.names = FALSE
            )
          }
        )
      }
      plots <- list()
      if (length(dat) > 0) {
        if (input$color_hist & input$combine_hist) {
          p <- dat[
            ,
            get_freq_table(
              I,
              nbins = input$histbins,
              normalize = "area"
            ),
            name
          ] %>%
          .[
            ,
            smooth_data(
              x,
              frequency,
              eval(input$histbins * input$smooth_factor)
            ),
            name
          ] %>%
          .[, frequency := y] %>%
          draw_hist(name_column = "name")
          box <- dat[
            name != "Overall",
            get_freq_table(
              I,
              nbins = input$histbins
            ),
            name
          ] %>%
          .[, summarize_freq_table(data.table(x, frequency)), name] %>%
          draw_boxplot(name_column = "name", show_axis = FALSE)
          plots[[1]] <- plot_grid(
            box,
            p,
            ncol = 1,
            axis = "tblr",
            align = "v",
            rel_heights = c(ifelse(input$show_boxplot, 0.2, 0), 0.8)
          )
        } else {
          for (i in seq_len(length(unique(dat$name)))) {
            p <- dat[name == unique(dat$name)[i], I] %>%
            get_freq_table(nbins = input$histbins, normalize = "area") %>%
            .[
              ,
              smooth_data(
                x,
                frequency,
                eval(input$histbins * input$smooth_factor)
              )
            ] %>%
            .[, frequency := y] %>%
            draw_hist(
              peaks = ifelse(
                input$combine_hist & !input$color_hist,
                input$npeaks,
                0
              )
            )
            box <- dat[name == unique(dat$name)[i], I] %>%
            get_freq_table(nbins = input$histbins) %>%
            summarize_freq_table() %>%
            draw_boxplot(show_axis = FALSE)
            plots[[i]] <- plot_grid(
              box,
              p,
              ncol = 1,
              axis = "tblr",
              align = "v",
              rel_heights = c(ifelse(input$show_boxplot, 0.2, 0), 0.8)
            )
          }
        }
      }
      if (length(plots) > 0) {
        plot_grid(plotlist = plots, nrow = 1)
      }
    }
  })
}
render_xyplot <- function(input, output, data_fun) {
  output$xyplot <- renderPlot({
    dat <- data_fun()
    if (input$show_scatterplot) {
      plots <- list()
      if (length(dat) > 0) {
        export_data <- data.table()
        for (i in seq_len(length(dat))) {
          x_stream <- input[[paste0("xstream", i)]]
          y_stream <- input[[paste0("ystream", i)]]
          plot_data <- dat[[i]] %>%
          resample(factor = input$resample) %>%
          .[, .(x = I.x, y = I.y)] %>%
          predict_joint_density(input$densitybins)
          export_data <- rbind(
            export_data,
            cbind(plot_data, name = paste0("dataset", i))
          )
          if (input$add_density) {
            plots[[i]] <- plot_data %>%
            draw_scatter(
              normalize = input$xy_normalize,
              density_column = "density",
              box = input$densitybox
            ) +
            labs(
              x = paste(x_stream, input[[x_stream]]$name),
              y = paste(y_stream, input[[y_stream]]$name)
            )
          } else {
            plots[[i]] <- plot_data %>%
            draw_scatter(normalize = input$xy_normalize) +
            labs(
              x = paste(x_stream, input[[x_stream]]$name),
              y = paste(y_stream, input[[y_stream]]$name)
            )
          }
        }
        output$xydata <- downloadHandler(
          filename = "xydata.csv",
          content = function(file) {
            write.csv(
              export_data,
              file,
              row.names = FALSE
            )
          }
        )
      }
      if (length(plots) > 0) {
        plot_grid(plotlist = plots, align = "hv", axis = "tblr")
      }
    }
  })
}

server <- function(input, output) {
  render_files(input, output)
  render_xy_selectors(input, output)

  full_data <- reactive({
    dat <- data.table()
    for (i in seq_len(input$nfiles)) {
      file_id <- paste0("file", i)
      if (!is.null(input[[file_id]])) {
        file_dat <- input[[file_id]]$datapath %>%
        read_tif(
          exclude_zero = FALSE,
          include_xy = TRUE
        ) %>%
        .[, name := paste(file_id, input[[file_id]]$name)]
        dat <- dat %>%
        rbind(file_dat)
      }
    }
    dat
  })
  corr_data <- reactive({
    dat <- full_data()
    data_list <- list()
    if (length(dat) > 0) {
      for (i in seq_len(input$nfiles - 1)) {
        x_stream <- input[[paste0("xstream", i)]]
        y_stream <- input[[paste0("ystream", i)]]
        x_dat <- dat[name == paste(x_stream, input[[x_stream]]$name)]
        y_dat <- dat[name == paste(y_stream, input[[y_stream]]$name)]
        data_list[[i]] <- join_tif(x_dat, y_dat, input$exclude_zero)
      }
    }
    data_list
  })

  observe({
    dat <- full_data()
    if (length(dat) > 0) {
      dat <- dat[, .(I, name)] %>%
      iff(input$exclude_zero, .[I > 0]) %>%
      iff(
        input$combine_hist,
        .[, .(I = c(I, I), name = c(name, rep("Overall", .N)))]
      ) %>%
      iff(
        (input$color_hist & !input$show_overall),
        .[name != "Overall"]
      ) %>%
      iff(
        (!input$color_hist & input$combine_hist),
        .[name == "Overall"]
      ) %>%
      .[
        ,
        get_freq_table(
          I,
          nbins = input$histbins,
          normalize = "area"
        ),
        name
      ]
      output$histogramdata <- downloadHandler(
        filename = "histogram_data.csv",
        content = function(file) {
          write.csv(dat, file, row.names = FALSE)
        }
      )
      output$boxdata <- downloadHandler(
        filename = "boxplot_data.csv",
        content = function(file) {
          write.csv(
            dat[, summarize_freq_table(data.table(x, frequency)), name],
            file,
            row.names = FALSE
          )
        }
      )
    }
  })

  render_preview(input, output, full_data)
  render_hist(input, output, full_data)
  render_xyplot(input, output, corr_data)
}